"{ Package: 'stx:goodies/refactoryBrowser/parser' }"

"{ NameSpace: Smalltalk }"

RBValueNode subclass:#RBMessageNode
	instanceVariableNames:'receiver selector selectorParts arguments'
	classVariableNames:''
	poolDictionaries:''
	category:'Refactory-AST-Nodes'
!

RBMessageNode comment:'BRMessageNode is an AST node that represents a message send.

Instance Variables:
	arguments       <SequenceableCollection of: BRValueNode>        our argument nodes
	receiver        <BRValueNode>   the receiver''s node
	selector        <Symbol | nil>  the selector we''re sending (cached)
	selectorParts   <SequenceableCollection of: BRValueToken>       the tokens for each keyword

'
!


!RBMessageNode class methodsFor:'instance creation'!

receiver: aValueNode selector: aSymbol 
	^self 
		receiver: aValueNode
		selector: aSymbol
		arguments: #()
!

receiver: aValueNode selector: aSymbol arguments: valueNodes 
	^(self new)
		receiver: aValueNode;
		arguments: valueNodes;
		selector: aSymbol;
		yourself
!

receiver: aValueNode selectorParts: keywordTokens arguments: valueNodes 
        ^((keywordTokens contains: [:each | each isPatternVariable]) 
                ifTrue: [RBPatternMessageNode] 
                ifFalse: [RBMessageNode]
        ) new 
            receiver: aValueNode
            selectorParts: keywordTokens
            arguments: valueNodes

    "Modified: / 24-05-2018 / 20:53:18 / Claus Gittinger"
! !

!RBMessageNode methodsFor:'accessing'!

allMessageNodes
        ^(super allMessageNodes)
                add: self;
                yourself
!

arg1
    ^ self arguments first
!

arg1: aNode 
        arguments := arguments copy.
        arguments at:1 put:aNode.
        aNode parent: self
!

arg2
    ^ self arguments at:2

    "Created: / 29-11-2017 / 12:53:49 / cg"
!

arg3
    ^ self arguments at:3

    "Created: / 29-11-2017 / 12:53:58 / cg"
!

arg:n
    ^ self arguments at:n

    "Created: / 29-11-2017 / 12:54:20 / cg"
!

argumentCount
    self assert:(self selector argumentCount == arguments size).    
    ^ arguments size

    "Created: / 06-02-2017 / 12:45:34 / cg"
    "Modified: / 06-02-2017 / 14:30:46 / cg"
!

arguments
        arguments isNil ifTrue: [^ #()].
        ^ arguments
!

arguments: argCollection 
        (arguments := argCollection) notNil ifTrue:[
            arguments do: [:each | each parent: self]
        ].
!

children
        ^(OrderedCollection with: receiver) 
            addAll: (arguments ? #());
            yourself
!

leftMostReceiver
    ^ receiver leftMostReceiver
!

precedence
	^self isUnary ifTrue: [1] ifFalse: [self isKeyword ifTrue: [3] ifFalse: [2]]
!

receiver
	^receiver
!

receiver: aValueNode 
	receiver := aValueNode.
	receiver parent: self
!

selector
    "notice: the selector is no longer always made a symbol,
     and the selector instVar is lazy.
     This was done to allow parsing huge amounts of code without automatically
     creating symbols"

    selector isNil ifTrue:[ 
        selector := self selectorString asSymbolIfInternedOrSelf.
    ].    
    ^selector

    "Modified (comment): / 06-02-2017 / 14:31:22 / cg"
!

selector: aSelector 
        | keywords numArgs |
        keywords := aSelector keywords.
        numArgs := aSelector argumentCount.
        numArgs == arguments size 
                ifFalse: 
                        [self 
                                error: 'Attempting to assign selector with wrong number of arguments.'].
        selectorParts := numArgs == 0 
                                ifTrue: [Array with: (RBIdentifierToken value: keywords first start: nil)]
                                ifFalse: 
                                        [keywords first last == $: 
                                                ifTrue: [keywords collect: [:each | RBKeywordToken value: each start: nil]]
                                                ifFalse: [Array with: (RBBinarySelectorToken value: aSelector start: nil)]].
        "/ self assert:(selector isNil or:[selector partsIfSelector asArray = keywords]).
        selector := aSelector asSymbolIfInternedOrSelf

    "Modified: / 08-02-2017 / 13:21:55 / stefan"
!

selectorString
        ^selector isNil
                ifTrue: [selector := self buildSelectorString asSymbolIfInternedOrSelf]
                ifFalse: [selector]

    "Modified: / 08-02-2017 / 13:22:29 / stefan"
!

sentMessages
    "warning: returns a string collection because we avoid creating symbols when parsing"
    
    ^(super sentMessages)
            add: self selector;
            yourself

    "Modified (comment): / 06-02-2017 / 14:35:11 / cg"
!

startWithoutParentheses
	^receiver start
!

stopWithoutParentheses
        ^arguments isEmptyOrNil 
                ifTrue: [selectorParts first stop]
                ifFalse: [arguments last stop]
! !

!RBMessageNode methodsFor:'comparing'!

= anObject 
        |a1 a2|

        self == anObject ifTrue: [^true].
        self class = anObject class ifFalse: [^false].
        (receiver = anObject receiver 
                and: [self selectorString = anObject selectorString]) ifFalse: [^false].
        1 to: arguments size
                do: [:i | 
                            a1 := (arguments at: i).
                            a2 := (anObject arguments at: i).
                            a1 = a2 ifFalse: [^false]].
        ^true
!

equalTo: anObject withMapping: aDictionary 
    "return true, if I am equal to another node, with a given variable name mapping.
     Side effect: update the mapping dictionary as required to reflect a matching mapping (if possible)"

    self class = anObject class ifFalse: [^false].
    self selector = anObject selector ifFalse: [^false]. 
    (receiver equalTo: anObject receiver withMapping: aDictionary) ifFalse: [^false].
    1 to: arguments size do: [:i | 
        ((arguments at: i) equalTo: (anObject arguments at: i) withMapping: aDictionary) ifFalse: [^false]
    ].
    ^true

    "Modified (comment): / 26-07-2020 / 17:52:17 / cg"
!

equalTo: anObject withSelectorMapping: aDictionary 
        "return true, if I am equal to another node, with a given selector name mapping.
         Side effect: update the mapping dictionary as required to reflect a matching mapping (if possible)"

        self class = anObject class ifFalse: [^false].
        (receiver equalTo: anObject receiver withSelectorMapping: aDictionary) ifFalse: [^false].
        (aDictionary at:self selector ifAbsentPut:[anObject selector]) = anObject selector ifFalse:[
            ^ false
        ].
        arguments size = (anObject arguments size) ifFalse: [^false].  
        1 to: arguments size
                do: 
                        [:i | 
                        ((arguments at: i) equalTo: (anObject arguments at: i)
                                withSelectorMapping: aDictionary) ifFalse: [^false]].
        ^true
!

hash
        ^(receiver hash bitXor: self selector hash)
                bitXor: (arguments isEmpty ifTrue: [0] ifFalse: [arguments first hash])
!

semanticallyEqualTo: aNode withMapping: aDictionary 
    "return true, if I am semantically equal to another node, with a given variable name mapping.
     Semantical equalTo: is defined slightly less strict, than equalTo:.
     Both check the parsetree, ignoring formatting and comments.
     However, in some situations, semanticallyEqualTo: detects the same intention,
     (which is mostly heuristic and builtin knowledge).
     For example: 
        foo isNil ifTrue:e1 ifFalse:e2 
     is semantcallyEqualTo:
        foo notNil ifTrue:e2 ifFalse:e1
     and also to:
        foo isNil ifFalse:e2 ifTrue:e1 
     The fallback implemented here is to ask for the regular equality.
     Side effect: update the mapping dictionary as required to reflect a matching mapping (if possible)"

    |mySelector otherSelector
     otherReceiver otherArguments
     trueExpr falseExpr  
     otherTrueExpr otherFalseExpr
     t idx|    
    
    self species = aNode species ifFalse: [^false].

    mySelector := self selectorString.
    otherSelector := aNode selectorString.

    otherReceiver := aNode receiver.
    otherArguments := aNode arguments.

    "/ the following stuff is unfinished and needs more work;
    "/ will extract the code into separate checking methods eventually.
    
    "/ check if-conditions
    (mySelector = #'ifTrue:ifFalse:') ifTrue:[
        trueExpr := (arguments at: 1).
        falseExpr := (arguments at: 2).
    ] ifFalse:[ (mySelector = #'ifFalse:ifTrue:') ifTrue:[
        falseExpr := (arguments at: 1).
        trueExpr := (arguments at: 2).
    ] ifFalse:[ (mySelector = #'ifTrue:') ifTrue:[
        trueExpr := (arguments at: 1).
        falseExpr := (RBLiteralNode literalToken:(RBLiteralToken value:nil))
    ] ifFalse:[ (mySelector = #'ifFalse:') ifTrue:[
        falseExpr := (arguments at: 1).
        trueExpr := (RBLiteralNode literalToken:(RBLiteralToken value:nil))
    ]]]].
    trueExpr notNil ifTrue:[
        (otherSelector = #'ifTrue:ifFalse:') ifTrue:[
            otherTrueExpr := (otherArguments at: 1).
            otherFalseExpr := (otherArguments at: 2).
        ] ifFalse:[ (otherSelector = #'ifFalse:ifTrue:') ifTrue:[
            otherFalseExpr := (otherArguments at: 1).
            otherTrueExpr := (otherArguments at: 2).
        ] ifFalse:[ (otherSelector = #'ifTrue:') ifTrue:[
            otherTrueExpr := (otherArguments at: 1).
            otherFalseExpr := (RBLiteralNode literalToken:(RBLiteralToken value:nil))
        ] ifFalse:[ (otherSelector = #'ifFalse:') ifTrue:[
            otherFalseExpr := (otherArguments at: 1).
            otherTrueExpr := (RBLiteralNode literalToken:(RBLiteralToken value:nil))
        ]]]].

        otherTrueExpr notNil ifTrue:[ 
            |condExpr otherCondExpr savedMappings| 

            condExpr := receiver.
            otherCondExpr := otherReceiver.

            savedMappings := aDictionary copy.
            (condExpr semanticallyEqualTo: otherCondExpr withMapping: aDictionary) ifFalse:[    
                "/ must undo any new mappings, since we retry the same match
                aDictionary clearContents; declareAllFrom:savedMappings.

                (condExpr semanticallyEqualToNegated:otherCondExpr withMapping: aDictionary) ifFalse:[
                    ^ false
                ].
                
                "/ cond is negation of each other; exchange true- and false expressions
                t := trueExpr.
                trueExpr := falseExpr.
                falseExpr := t.
            ].
            
            (trueExpr semanticallyEqualTo:otherTrueExpr withMapping: aDictionary) ifFalse: [^false].
            (falseExpr semanticallyEqualTo:otherFalseExpr withMapping: aDictionary) ifFalse: [^false].
            ^ true.        
        ].
    ].

    "/ relational operations
    
    "/ same relOp with args exchanged?
    "/ i.e. '(a > b)' vs. '(b < 0)' 
    ((idx := #('<' '<=' '>' '>=') indexOf:mySelector) ~~ 0) ifTrue:[
        (otherSelector = (#('>' '>=' '<' '<=') at:idx)) ifTrue:[
            (receiver semanticallyEqualTo:otherArguments first withMapping: aDictionary) ifFalse: [^ false].
            (arguments first semanticallyEqualTo:otherReceiver withMapping: aDictionary) ifFalse: [^ false].
            ^ true
        ].    
    ].
    
    mySelector = 'not' ifTrue:[
        |savedMappings|
        
        savedMappings := aDictionary copy.
        (receiver semanticallyEqualToNegated:aNode withMapping:aDictionary) ifTrue:[^ true].
        "/ must undo any new mappings, since we retry the same match
        aDictionary clearContents; declareAllFrom:savedMappings.
    ].
    
    (otherSelector = 'not') ifTrue:[
        |savedMappings|

        savedMappings := aDictionary copy.
        (self semanticallyEqualToNegated:aNode receiver withMapping:aDictionary) ifTrue:[^ true].
        "/ must undo any new mappings, since we retry the same match
        aDictionary clearContents; declareAllFrom:savedMappings.
    ].

    "/ 'a isNil' vs. 'a == nil'    
    (((mySelector = 'isNil') and:[ (otherSelector = '==') ])
      or:[ ((mySelector = 'notNil') and:[ (otherSelector = '~~') ]) ]
    ) ifTrue:[
        |op1 otherOp1 savedMappings|

        otherReceiver isLiteralNil ifTrue:[
            otherOp1 := otherArguments first
        ] ifFalse:[
            otherArguments first isLiteralNil ifTrue:[
                otherOp1 := otherReceiver
            ] ifFalse:[
                ^ false
            ].
        ].
        (receiver semanticallyEqualTo:otherOp1 withMapping:aDictionary) ifTrue:[^ true].
        ^ false.
    ].
    "/ 'a isNil' vs. 'a == nil'    
    (((otherSelector = 'isNil') and:[ (mySelector = '==') ])
      or:[ ((otherSelector = 'notNil') and:[ (mySelector = '~~') ]) ]
    ) ifTrue:[
        |op1 myOp1 savedMappings|

        receiver isLiteralNil ifTrue:[
            myOp1 := arguments first
        ] ifFalse:[
            arguments first isLiteralNil ifTrue:[
                myOp1 := receiver
            ] ifFalse:[
                ^ false
            ].
        ].
        (otherReceiver semanticallyEqualTo:myOp1 withMapping:aDictionary) ifTrue:[^ true].
        ^ false.
    ].


    "/ normal checks    
    (mySelector = otherSelector) ifFalse: [ ^ false].
    (receiver semanticallyEqualTo: aNode receiver withMapping: aDictionary) ifFalse: [^ false].

    arguments with:otherArguments do:[:myArgI :otherArgI | 
        (myArgI semanticallyEqualTo: otherArgI withMapping: aDictionary) ifFalse: [^false]
    ].
    ^true

    "Modified: / 06-02-2017 / 15:02:24 / cg"
!

semanticallyEqualToNegated:expr2 withMapping:mappingDict
    "/ return true,
    "/ if the receiver and expr2 compute the negated result of each other;
    "/ eg. '(a > b)' vs. '(a <= b)'
    "/ or  '(a > b) not' vs. '(a > b)'

    |mySelector savMapping idx|

    mySelector := self selector.
    (mySelector = 'not') ifTrue:[
        savMapping := mappingDict copy.
        (receiver semanticallyEqualTo:expr2 withMapping:mappingDict) ifTrue:[^ true].
        "/ undo mappings.
        mappingDict clearContents; declareAllFrom:savMapping.
    ] ifFalse:[
        (expr2 isMessage and:[expr2 selector = 'not']) ifTrue:[
            savMapping := mappingDict copy.
            (self semanticallyEqualTo:expr2 receiver withMapping:mappingDict) ifTrue:[^ true].
            "/ undo mappings.
            mappingDict clearContents; declareAllFrom:savMapping.
        ]
    ].

    ((mySelector = 'notNil')
    and:[ (expr2 isMessage and:[expr2 selector = 'isNil'])]) ifTrue:[
        savMapping := mappingDict copy.
        (receiver semanticallyEqualTo:expr2 receiver withMapping:mappingDict) ifTrue:[^ true].
        "/ undo mappings.
        mappingDict clearContents; declareAllFrom:savMapping.
    ].
    ((mySelector = 'isNil')
    and:[ (expr2 isMessage and:[expr2 selector = 'notNil'])]) ifTrue:[
        savMapping := mappingDict copy.
        (receiver semanticallyEqualTo:expr2 receiver withMapping:mappingDict) ifTrue:[^ true].
        "/ undo mappings.
        mappingDict clearContents; declareAllFrom:savMapping.
    ].
    
    ((self argumentCount == 1)
      and:[ (expr2 isMessage)
      and:[ (expr2 argumentCount == 1) ]]
    ) ifTrue:[
        |sel1 sel2 op1 op2 otherOp1 otherOp2|

        sel1 := mySelector.
        sel2 := expr2 selector.
        op1 := receiver.
        op2 := arguments first.
        otherOp1 := expr2 receiver.
        otherOp2 := expr2 arguments first.
        
        ((idx := #('<' '<=' '>' '>=') indexOf:sel1) ~~ 0) ifTrue:[
            (sel2 = (#('>=' '>' '<=' '<') at:idx)) ifTrue:[
                savMapping := mappingDict copy.
                (op1 semanticallyEqualTo:otherOp1 withMapping:mappingDict) ifTrue:[
                    (op2 semanticallyEqualTo:otherOp2 withMapping:mappingDict) ifTrue:[^ true].
                ].
                "/ undo mappings.
                mappingDict clearContents; declareAllFrom:savMapping.
            ].
            ^ false
        ].
            
        ((idx := #('=' '==' '~=' '~~') indexOf:sel1) ~~ 0) ifTrue:[
            (sel2 = (#('~=' '~~' '=' '==') at:idx)) ifTrue:[
                savMapping := mappingDict copy.
                (op1 semanticallyEqualTo:otherOp1 withMapping:mappingDict) ifTrue:[
                    (op2 semanticallyEqualTo:otherOp2 withMapping:mappingDict) ifTrue:[^ true].
                ].
                "/ undo mappings.
                mappingDict clearContents; declareAllFrom:savMapping.

                "/ in reality, reversed argument are allowed for '=' and '~='
                "/ however, these operations might not be communative (although they should).
                "/ therefore, we only try the inlined identity comparisons
                ((sel1 = '==') or:[sel1 = '~~']) ifTrue:[
                    (op1 semanticallyEqualTo:otherOp2 withMapping:mappingDict) ifTrue:[
                        (op2 semanticallyEqualTo:otherOp1 withMapping:mappingDict) ifTrue:[^ true].
                    ].
                    "/ undo mappings.
                    mappingDict clearContents; declareAllFrom:savMapping.
                ].
            ].
        ]    
    ].
    ^ false

    "Created: / 06-02-2017 / 13:33:14 / cg"
! !

!RBMessageNode methodsFor:'copying-private'!

postCopy
        super postCopy.
        self receiver: receiver copy.
        self arguments: (arguments collect: [:each | each copy])

    "Modified: / 18-06-2015 / 20:33:45 / Jan Vrany <jan.vrany@fit.cvut.cz>"
! !

!RBMessageNode methodsFor:'enumeration'!

allMessageNodesDo:aBlock
    receiver allMessageNodesDo:aBlock.
    arguments do:[:eachArg |
        eachArg allMessageNodesDo:aBlock
    ].    
    aBlock value:self.
!

allNonPatternMessageNodesDo:aBlock
    self isPatternNode ifFalse:[
        receiver allNonPatternMessageNodesDo:aBlock.
        arguments do:[:eachArg |
            eachArg allNonPatternMessageNodesDo:aBlock
        ].    
        aBlock value:self.
    ].

    "Created: / 16-11-2016 / 18:19:22 / cg"
!

variableNodesDo:aBlock
    "evaluate a Block for all variable nodes"

    self isPatternNode ifFalse:[
        receiver variableNodesDo:aBlock.
        arguments do:[:each | each variableNodesDo:aBlock].
    ]

    "Modified: / 16-11-2016 / 18:44:36 / cg"
! !

!RBMessageNode methodsFor:'initialization & release'!

receiver: aValueNode selectorParts: keywordTokens arguments: valueNodes 
        self receiver: aValueNode.
        selectorParts := keywordTokens.
        valueNodes notNil ifTrue:[
            self arguments: valueNodes.
        ].
! !

!RBMessageNode methodsFor:'libtool3 support'!

highlightInCodeView: aCodeView

    ^aCodeView highlightMessageNode: self

    "Created: / 18-02-2008 / 17:51:11 / janfrog"
!

leftClickMenuInCodeView: aCodeView 
    ^ aCodeView leftClickMenuForMessageNode: self.

    "Created: / 18-02-2008 / 19:04:45 / janfrog"
!

middleClickMenuInCodeView: aCodeView 
    ^ aCodeView middleClickMenuForMessageNode: self.

    "Created: / 18-02-2008 / 19:04:58 / janfrog"
!

rightClickMenuInCodeView: aCodeView 
    ^ aCodeView rightClickMenuForMessageNode: self.

    "Created: / 18-02-2008 / 19:05:18 / janfrog"
! !

!RBMessageNode methodsFor:'matching'!

copyInContext: aDictionary 
	^(self class new) receiver: (receiver copyInContext: aDictionary);
		selectorParts: (selectorParts collect: [:each | each removePositions]);
		arguments: (arguments collect: [:each | each copyInContext: aDictionary]);
		yourself
!

match: aNode inContext: aDictionary 
    aNode species == self species ifFalse: [^false].
    self selectorString = aNode selectorString ifFalse: [^false].
    (receiver match: aNode receiver inContext: aDictionary) ifFalse: [^false].
    1 to: arguments size do: [:i | 
        ((arguments at: i) match: (aNode arguments at: i) inContext: aDictionary)
            ifFalse: [^false]
    ].
    ^ true

    "Modified (format): / 19-11-2016 / 16:01:23 / cg"
! !

!RBMessageNode methodsFor:'private'!

buildSelector
    "concatenate the selector parts, and intern as a symbol"

    "notice: the selector is no longer always made a symbol.
     This was done to allow parsing huge amounts of code without automatically
     creating symbols"

    ^ self buildSelectorString "/ asSymbol

    "Modified (comment): / 06-02-2017 / 13:02:23 / cg"
!

buildSelectorString
        "concatenate the selector parts, but do not intern"

        | selectorStream |
        selectorStream := WriteStream on: (String new: 50).
        selectorParts do: [:each | selectorStream nextPutAll: each value].
        ^selectorStream contents
!

selectorParts
	^selectorParts
!

selectorParts: tokenCollection 
	selectorParts := tokenCollection
! !

!RBMessageNode methodsFor:'querying'!

bestNodeFor: anInterval 
	(self intersectsInterval: anInterval) ifFalse: [^nil].
	(self containedBy: anInterval) ifTrue: [^self].
	selectorParts do: 
			[:each | 
			((anInterval first between: each start and: each stop) 
				or: [each start between: anInterval first and: anInterval last]) 
					ifTrue: [^self]].
	self children do: 
			[:each | 
			| node |
			node := each bestNodeFor: anInterval.
			node notNil ifTrue: [^node]]
!

firstLineNumber
    ^ receiver firstLineNumber
!

lastLineNumber
    arguments notEmptyOrNil ifTrue:[
        ^ arguments last lastLineNumber
    ].
    ^ selectorParts last lineNumber.
"/    ^ lineNumber.
! !

!RBMessageNode methodsFor:'replacing'!

replaceNode: aNode withNode: anotherNode 
        "If we're inside a cascade node and are changing the receiver, change all the receiver's"

        receiver == aNode 
                ifTrue: 
                        [self receiver: anotherNode.
                        (parent notNil and: [parent isCascade]) 
                                ifTrue: [parent messages do: [:each | each receiver: anotherNode]]].
        self arguments: (arguments 
                                collect: [:each | each == aNode ifTrue: [anotherNode] ifFalse: [each]])
! !

!RBMessageNode methodsFor:'testing'!

isBinary
	^(self isUnary or: [self isKeyword]) not
!

isCascaded
	^parent notNil and: [parent isCascade]
!

isFirstCascaded
	^self isCascaded and: [parent messages first == self]
!

isKeyword
	^selectorParts first value last == $:
!

isKeywordMessage
     ^ self isKeyword      
!

isLoop
    |sel|

    sel := self selectorString.
    
    (sel = #loop) ifTrue:[^ true].
    ((sel = #whileTrue:) and:[ self receiver hasTrueValue ]) ifTrue:[^ true].
    ((sel = #whileFalse:) and:[ self receiver hasFalseValue ]) ifTrue:[^ true].

    ^ false

    "Created: / 16-07-2017 / 13:42:08 / cg"
!

isMessage
	^true
!

isMessageWithSelector:aStringOrSymbol
        ^ self selectorString = aStringOrSymbol
!

isUnary
        ^arguments isEmptyOrNil
!

lastIsReturn
    ^ (#(#ifTrue:ifFalse: #ifFalse:ifTrue:) includes: self selectorString) 
        and:[arguments first isBlock 
        and:[arguments first body lastIsReturn
        and:[arguments last isBlock 
        and: [arguments last body lastIsReturn]]]]
!

testLoop1
    [
        Transcript showCR:'looping...'
    ] loop

    "
     self assert:( (RBMessageNode compiledMethodAt:#testLoop1) parseTree lastIsLoop )
     self assert:( (RBMessageNode compiledMethodAt:#testLoop2) parseTree lastIsLoop )
     self assert:( (RBMessageNode compiledMethodAt:#testLoop3) parseTree lastIsLoop )
    "

    "Created: / 16-07-2017 / 14:19:27 / cg"
!

testLoop2
    [true] whileTrue:[
        Transcript showCR:'looping...'
    ]

    "
     self assert:( (RBMessageNode compiledMethodAt:#testLoop1) parseTree lastIsLoop )
     self assert:( (RBMessageNode compiledMethodAt:#testLoop2) parseTree lastIsLoop )
     self assert:( (RBMessageNode compiledMethodAt:#testLoop3) parseTree lastIsLoop )
    "

    "Created: / 16-07-2017 / 14:19:40 / cg"
!

testLoop3
    [false] whileFalse:[
        Transcript showCR:'looping...'
    ]

    "
     self assert:( (RBMessageNode compiledMethodAt:#testLoop1) parseTree lastIsLoop )
     self assert:( (RBMessageNode compiledMethodAt:#testLoop2) parseTree lastIsLoop )
     self assert:( (RBMessageNode compiledMethodAt:#testLoop3) parseTree lastIsLoop )
    "

    "Created: / 16-07-2017 / 14:19:47 / cg"
! !

!RBMessageNode methodsFor:'visitor'!

acceptVisitor: aProgramNodeVisitor 
	^aProgramNodeVisitor acceptMessageNode: self
! !

!RBMessageNode class methodsFor:'documentation'!

version
    ^ '$Header$'
!

version_CVS
    ^ '$Header$'
! !

